//==============================================================================
//=  Ubuntu Linux 环境信息学评测系统
//=  最后编辑时间：2009年03月30日14:22
//==============================================================================

//******************************************************************************
//*  宏定义
#define BUFSIZE 2000

//******************************************************************************
//*  引入头文件
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>

//******************************************************************************
//*  全局类型
typedef struct t_PROB 	//题目链表结点
{
	char* Name; 					//题目名称
	char* Plugin;					//评测插件
	int Count; 						//测试点数量
	int* Score; 					//每个测试点分数
	int* Limit;						//每个测试点时限
	struct t_PROB* Next;	//链表指针
}PROB;

typedef struct t_USER		//选手链表结点
{
	char* Name;						//用户名
	PROB* Prob;						//题目链表结点
	struct t_USER* Next;		//链表指针
}USER;

//******************************************************************************
//*  全局变量
char* DataDir; 		//测试数据根目录
char* SrcDir; 		//选手程序根目录
char* Result;			//评测结果输出文件名
PROB* hProb;			//题目链表头结点
USER* hUser;			//选手链表头结点

//******************************************************************************
//*  系统初始化

//从文件读取一行
char* fGetLine(FILE* fin)
{
	char buf[BUFSIZE];
	char* r;
	if(fgets(buf, BUFSIZE, fin)!=NULL)
	{
		buf[strlen(buf)-1]='\0';
		r=malloc((strlen(buf)+1)*sizeof(char));
		strcpy(r, buf);
		return r;
	}
	else
		return NULL;
}

//初始化
int InitSystem(char* fn)
{
	FILE* fin;
	char* buf;
	USER* utmp;
	PROB* ptmp;
	PROB* up;
	int i, j;
	
	//打开配置文件
	if(fn==NULL)
		fin=fopen("OIJudge.cfg", "rt");
	else
		fin=fopen(fn, "rt");
		
	if(fin==NULL)
		return -1;
	
	//读取测试数据根目录
	buf=fGetLine(fin);
	while(buf[0]=='#' || buf[0]=='\0')
	{
		free(buf);
		buf=fGetLine(fin);
	}
	DataDir=buf;
	
	//读取选手程序根目录
	buf=fGetLine(fin);
	while(buf[0]=='#' || buf[0]=='\0')
	{
		free(buf);
		buf=fGetLine(fin);
	}
	SrcDir=buf;
	
	//读取评测结果输出文件名
	buf=fGetLine(fin);
	while(buf[0]=='#' || buf[0]=='\0')
	{
		free(buf);
		buf=fGetLine(fin);
	}
	Result=buf;
	
	//读取选手列表
	buf=fGetLine(fin);
	while(buf[0]=='#' || buf[0]=='\0')
	{
		free(buf);
		buf=fGetLine(fin);
	}
	hUser=malloc(sizeof(USER));
	memset(hUser, 0, sizeof(USER));
	utmp=hUser;
	while(buf[0]!='\0')
	{
		utmp->Next=malloc(sizeof(USER));
		utmp=utmp->Next;
		memset(utmp, 0, sizeof(USER));
		utmp->Name=buf;
		buf=fGetLine(fin);
	}
	
	//读取题目信息
	buf=fGetLine(fin);
	while(buf[0]=='#' || buf[0]=='\0')
	{
		free(buf);
		buf=fGetLine(fin);
	}
	hProb=malloc(sizeof(PROB));
	memset(hProb, 0, sizeof(PROB));
	ptmp=hProb;
	while(buf!=NULL && buf[0]!='\0')	
	{
		ptmp->Next=malloc(sizeof(PROB));
		ptmp=ptmp->Next;
		memset(ptmp, 0, sizeof(PROB));
		
		//分解题目名称
		for(i=0;buf[i]!=' ';i++);
		buf[i]='\0';
		ptmp->Name=malloc((i+1)*sizeof(char));
		strcpy(ptmp->Name, buf);
		
		//分解测试点数量
		ptmp->Count=0;
		for(i++;buf[i]!=' ';i++)
			ptmp->Count=ptmp->Count*10+buf[i]-'0';
		
		//分解每点得分
		ptmp->Score=malloc(ptmp->Count*sizeof(int));
		for(j=0;j<ptmp->Count;j++)
		{
			ptmp->Score[j]=0;
			for(i++;buf[i]!=' ';i++)
				ptmp->Score[j]=ptmp->Score[j]*10+buf[i]-'0';
		}
		
		//分解每点时限
		ptmp->Limit=malloc(ptmp->Count*sizeof(int));
		for(j=0;j<ptmp->Count;j++)
		{
			ptmp->Limit[j]=0;
			for(i++;buf[i]!=' ' && buf[i]!='\0';i++)
				ptmp->Limit[j]=ptmp->Limit[j]*10+buf[i]-'0';
		}
				
		//读取评测插件
		free(buf);
		buf=fGetLine(fin);
		ptmp->Plugin=buf;
		
		buf=fGetLine(fin);
	}
	
	//构建选手信息数据结构
	utmp=hUser->Next;
	while(utmp!=NULL)
	{
		utmp->Prob=malloc(sizeof(PROB));
		memset(utmp->Prob, 0, sizeof(PROB));
		up=utmp->Prob;
		ptmp=hProb->Next;
		while(ptmp!=NULL)
		{
			up->Next=malloc(sizeof(PROB));
			up=up->Next;
			memset(up, 0, sizeof(PROB));
			
			up->Name=malloc((strlen(ptmp->Name)+1)*sizeof(char));
			strcpy(up->Name, ptmp->Name);
			up->Count=ptmp->Count;
			up->Limit=malloc(up->Count*sizeof(int));
			up->Score=malloc(up->Count*sizeof(int));
			
			ptmp=ptmp->Next;
		}
		
		utmp=utmp->Next;
	}
	
	//关闭文件配置文件
	fclose(fin);
	
	return 0;
}

//******************************************************************************
//* 清理内存
void Clear()
{
	PROB* ptmp;
	USER* utmp;
	
	//清理数据根目录
	free(DataDir);
	
	//清理选手程序根目录
	free(SrcDir);
	
	//清理题目信息链表	
	while(hProb!=NULL)
	{
		ptmp=hProb->Next;
		free(hProb->Name);
		free(hProb->Score);
		free(hProb->Limit);
		free(hProb->Plugin);
		free(hProb);
		hProb=ptmp;
	}
	
	//清理选手信息链表
	while(hUser!=NULL)
	{
		utmp=hUser->Next;
		free(hUser->Name);
		while(hUser->Prob)
		{
			ptmp=hUser->Prob->Next;
			free(hUser->Prob->Name);
			free(hUser->Prob->Score);
			free(hUser->Prob->Limit);
			free(hUser->Prob);
			hUser->Prob=ptmp;
		}		
		hUser=utmp;
	}
}

//******************************************************************************
//*  评测系统核心
void Judge()
{
	USER* u;
	PROB* p;
	PROB* up;
	char s[BUFSIZE], arg1[BUFSIZE], arg2[BUFSIZE], arg3[BUFSIZE];
	int i, status;
	pid_t wpid, rpid, tpid;
	
	//对每位选手进行评测
	u=hUser->Next;
	while(u!=NULL)
	{
		printf("---------------------------------------\n");
		printf("%s\n", u->Name);
		up=u->Prob->Next;
		
		//进入到选手文件夹
		sprintf(s, "%s/%s/", SrcDir, u->Name);
		if(chdir(s)==0)
		{
			//对每位选手的每个题目进行评测
			p=hProb->Next;
			i=0;
			while(p!=NULL)
			{
				printf("\n\t*** %s ***\n", p->Name);
				//进入到选手题目文件夹
				sprintf(s, "%s/%s/%s/", SrcDir, u->Name, p->Name);
				if(chdir(s)==0)
				{
					//编译选手程序
					sprintf(s, "rm %s -f", p->Name);//删除可执行文件
					system(s);
					sprintf(s, "%s.cpp", p->Name);
					if(access(s, 0)==0)//选手提交C++代码
					{
						sprintf(s, "g++ %s.cpp -o %s", p->Name, p->Name);
						system(s);
					}
					else
					{
						sprintf(s, "%s.c", p->Name);
						if(access(s, 0)==0)//选手c代码
						{
							sprintf(s, "gcc %s.c -o %s", p->Name, p->Name);
							system(s);
						}
						else
							up->Score[0]=-3;//选手代码未提交
					}
					
					//评测得分
					if(access(p->Name, 0)==0)
					{
						system("rm *.out -f");
						for(i=0;i<p->Count;i++)
						{
							//删除输入、输出文件
							system("rm *.in -f");							
							//拷贝标准输入
							sprintf(s, "cp %s/%s/%s%d.in %s.in", DataDir, p->Name, p->Name, i, p->Name);
							system(s);
							
							//运行选手程序
							sprintf(s, "./%s", p->Name);
							wpid=fork();
							if(wpid==0)
							{
								usleep(p->Limit[i]);
								return;
							}
							rpid=fork();
							if(rpid==0)
								execl(s, s, NULL);
							tpid=wait(NULL);
							while(tpid!=wpid && tpid!=rpid)
								tpid=wait(NULL);
							kill(wpid, SIGKILL);
							kill(rpid, SIGKILL);
							if(tpid==wpid)
							{
								printf("\t\t测试点%d: 超时\n", i);
								up->Score[i]=-5;//超时
								
								//删除错误的输出文件
								sprintf(s, "rm %s.out -f", p->Name);
								system(s);
							}
							else
							{
								rpid=fork();
								if(rpid==0)
								{
									sprintf(arg1, "%s/%s/%s%d.out", DataDir, p->Name, p->Name, i);
									sprintf(arg2, "%s.out", p->Name);
									sprintf(arg3, "%d", p->Score[i]);
									execl(p->Plugin, p->Plugin, arg1, arg2, arg3, NULL);
								}
								waitpid(rpid, &status, 0);
								up->Score[i]=WEXITSTATUS(status);
								if(up->Score[i]==0)
									printf("\t\t测试点%d: 输出错误！\n", i);
								else
									printf("\t\t测试点%d: %d\n", i, up->Score[i]);
								
								//保存选手输出
								sprintf(s, "mv %s.out %s%d.out", p->Name, p->Name, i);
								system(s);							
							}
						}
						system("rm *.in -f");
					}
					else if (up->Score[0]==-3)
						printf("\t\t代码文件未提交！\n");
					else
					{
						printf("\t\t编译错误！\n");
						up->Score[0]=-4;//编译错误
					}
				}
				else
				{
					printf("\t\t题目文件夹不存在！\n");
					up->Score[0]=-2;//题目文件夹不存在				
				}
				
				p=p->Next;
				up=up->Next;
			}
		}
		else
		{
			printf("\t选手提交文件夹不存在！\n");
			up->Score[0]=-1;//选手提交文件夹不存在
		}
		
		u=u->Next;
	}
	printf("---------------------------------------\n");
}

//******************************************************************************
//*  输出评测结果
void Write()
{
	FILE *fout;
	USER* u;
	PROB* p;
	int i, sum;
	
	fout=fopen(Result, "wt");
	u=hUser->Next;
	while(u!=NULL)
	{
		sum=0;
		fprintf(fout, "---------------------------------------------\n");
		fprintf(fout, "%s\n", u->Name);
		p=u->Prob->Next;
		if(p->Score[0]==-1)
			fprintf(fout, "\t选手文件夹不存在！\n");
		else
			while(p!=NULL)
			{
				fprintf(fout, "\t%s: ", p->Name);
				if(p->Score[0]==-2)
					fprintf(fout, "题目文件夹不存在！\n");
				else if(p->Score[0]==-3)
					fprintf(fout, "代码文件不存在！\n");
				else if(p->Score[0]==-4)
					fprintf(fout, "编译错误！\n");
				else
				{
					for(i=0;i<p->Count;i++)
					{
						if(p->Score[i]==-5)
							fprintf(fout, "超时");
						else
						{
							fprintf(fout, "%d", p->Score[i]);
							sum=sum+p->Score[i];
						}
						
						if(i!=p->Count-1)
							fprintf(fout, " ");
					}
					fprintf(fout, "\n");			
				}
				p=p->Next;
			}
		fprintf(fout, "总分: %d\n", sum);
		u=u->Next;
	}
	fprintf(fout, "---------------------------------------------\n");
	fclose(fout);
}

//******************************************************************************
//*  主函数
int main(int argc, char* argv[])
{
	char* fn;
	system("clear");
	printf("*************************************************************\n");
	printf("                Ubuntu Linux 信息学评测软件\n");
	printf("                    2009年3月18日编译\n");
	printf("*************************************************************\n");
	usleep(2000000);
	if(argc==2)
		fn=argv[1];
	else
		fn=NULL;
	if(InitSystem(fn)!=0)
	{
		printf("配置文件无法打开！\n");
		return -1;
	}
	Judge();
	Write();
	Clear();
	return 0;
}

//******************************************************************************
